@@ -25,7 +25,7 @@ use POE::Component::IKC::Specifier;
 require Exporter;
 @ISA = qw(Exporter);
 @EXPORT = qw(create_ikc_responder $ikc);
-$VERSION = '0.09';
+$VERSION = '0.10';
 
 sub DEBUG { 0 }
 
@@ -61,7 +61,7 @@ sub _start
 
 sub _stop
 {
-#    warn "$_[HEAP] responder _stop\n";
+#    warn "$$: $_[HEAP] responder _stop\n";
 }
 
 #----------------------------------------------------
@@ -136,7 +136,7 @@ sub remote_error
 {
     my($heap, $msg) = @_[HEAP, ARG0];
 
-    warn "Remote error: $msg\n";
+    warn "$$: Remote error: $msg\n";
 }
 
 ##############################################################################
@@ -325,7 +325,7 @@ sub request
     {
         chomp($@);
         my $err=$@.' ['.specifier_name($to).']';
-        DEBUG && warn "Error in request: $err\n";
+        DEBUG && warn "$$: Error in request: $err\n";
         unless($request->{is_error})    # don't send an error message back
         {                               # if this was an error itself
             $self->send_msg({ event=>$request->{errors_to},
@@ -333,7 +333,7 @@ sub request
                             });
         } else
         {
-            warn Dumper $request;
+            warn $$,  Dumper $request;
         }
     }
 }
@@ -487,8 +487,8 @@ sub send_msg
     unless(@channels)
     {
         warn (($name eq '*') 
-                  ? "Not connected to any foreign kernels.\n" 
-                  : "Unknown kernel '$name'.\n");
+                  ? "$$: Not connected to any foreign kernels.\n" 
+                  : "$$: Unknown kernel '$name'.\n");
         return 0;
     }
 
@@ -522,7 +522,7 @@ sub send_msg
     }
 
     DEBUG2 && print specifier_name($to), " sent to $count kernel(s).\n";
-    DEBUG and do {warn "send_msg failed!" unless $count};
+    DEBUG and do {warn "$$: send_msg failed!" unless $count};
     return $count;
 }
 
@@ -575,17 +575,17 @@ sub call
     {
         if($rsvp)
         {
-            warn "Bad 'rsvp' parameter '$rsvp' in poe:/IKC/call\n";
+            warn "$$: Bad 'rsvp' parameter '$rsvp' in poe:/IKC/call\n";
         } else
         {
-            warn "Missing 'rsvp' parameter in poe:/IKC/call\n";
+            warn "$$: Missing 'rsvp' parameter in poe:/IKC/call\n";
         }
         return;
     }
     $rsvp=$t;
     unless($rsvp->{state})
     {
-        warn "rsvp state not set in poe:/IKC/call\n";
+        warn "$$: rsvp state not set in poe:/IKC/call\n";
         return;
     }
 
@@ -777,7 +777,7 @@ sub _subscribe_receipt
 
     if(not $ses or not ref $ses)
     {
-        warn "Refused to subscribe to $spec\n";
+        warn "$$: Refused to subscribe to $spec\n";
         $accepted=0;
     } else
     {
@@ -912,7 +912,7 @@ sub _default
             @_[KERNEL, HEAP, SENDER, ARG0, ARG1];
     return if $state =~ /^_/;
 
-    warn "Attempt to respond to a foreign post with $state\n"
+    warn "$$: Attempt to respond to a foreign post with $state\n"
         if not $heap->{rsvp};
 
     $POE::Component::IKC::Responder::ikc->send_msg(
