@@ -27,10 +27,10 @@ require Exporter;
 
 @ISA = qw(Exporter);
 @EXPORT = qw(create_ikc_server);
-$VERSION = '0.09';
+$VERSION = '0.10';
 
 sub DEBUG { 0 }
-sub DEBUG_USR2 { 1 }
+sub DEBUG_USR2 { 0 }
 
 ###############################################################################
 #----------------------------------------------------
@@ -44,16 +44,12 @@ sub create_ikc_server
     $params{port}||=603;                # POE! (almost :)
     create_ikc_responder();
     POE::Session->new( 
-                    _start   => \&server_start,
-                    _stop    => \&server_stop,
-                    error    => \&server_error,
-                    'accept' => \&server_accept,
-                    'other_signal' => \&other_signal,
-                    'signal' => \&server_signal,
-                    'fork'   => \&server_fork,
-                    'retry'  => \&server_retry,
-                    'waste_time' => \&server_waste,
-                    'babysit' => \&server_babysit,
+                    __PACKAGE__, [qw(
+                        _start _stop error 
+                        accept fork retry waste_time 
+                        babysit rogues
+                        sig_CHLD sig_INT sig_other
+                        )],
                     [\%params],
                   );
 }
@@ -65,11 +61,13 @@ sub _select_define
     my($heap, $on)=@_;
 
     my $state;
+    my $err="possible redefinition of POE internals";
+    $err="failure to bind to port" if $POE::VERSION <= 0.1005;
     if($on) {
         $state=$heap->{wheel}->{'state_accept'};
         
         unless(defined $state) {
-            die "No 'state_accept' in $heap->{wheel}, possible redefinition of POE internals or failure to bind to port.\n";
+            die "No 'state_accept' in $heap->{wheel}, $err.\n";
         }
     }
     my $c=0;
@@ -78,7 +76,7 @@ sub _select_define
         $poe_kernel->select_read($heap->{wheel}->{$hndl}, $state);
         $c++;
     }
-    die "No socket_handle in $heap->{wheel}, possible redefinition of POE internals.\n" unless $c;
+    die "No socket_handle in $heap->{wheel}, $err.\n" unless $c;
     return;
 }
 
@@ -86,7 +84,7 @@ sub _select_define
 # Accept POE's standard _start event, and set up the listening socket
 # factory.
 
-sub server_start
+sub _start
 {
     my($heap, $params, $kernel) = @_[HEAP, ARG0, KERNEL];
 
@@ -108,27 +106,39 @@ sub server_start
     # Children put the state back in place after the fork
     _select_define($heap, 0);
 
-    $kernel->sig('CHLD', 'signal');
-    $kernel->sig('INT', 'signal');
-    DEBUG_USR2 and $kernel->sig('USR2', 'other_signal');
+    $kernel->sig('CHLD', 'sig_CHLD');
+    $kernel->sig('INT', 'sig_INT');
+    DEBUG_USR2 and $kernel->sig('USR2', 'sig_other');
+
                                         # keep track of children
     $heap->{children} = {};
     $heap->{'failed forks'} = 0;
     $heap->{verbose}=$params->{verbose}||0;
     $heap->{"max connections"}=$params->{connections}||1;
-                                        # change behavior for children
-    $heap->{'is a child'} = 0;
+                                        
+    $heap->{'is a child'} = 0;          # change behavior for children
     foreach (2..$params->{processes}) { # fork the initial set of children
         $kernel->yield('fork');
     }
     $kernel->yield('waste_time');    
-    $kernel->yield('babysit')           if $params->{babysit};    
+    if($params->{babysit}) {
+        $heap->{babysit}=$params->{babysit};
+        delete($heap->{"proctable"});
+        eval {
+            require Proc::ProcessTable;
+            $heap->{"proctable"}=new Proc::ProcessTable;
+        };
+        DEBUG and do {
+            print "Unable to load Proc::ProcessTable: $@\n" if $@;
+        };
+        $kernel->yield('babysit');
+    }
     return;
 }
 
 #------------------------------------------------------------------------------
 # This event keeps this POE kernel alive
-sub server_waste
+sub waste_time
 {
     my($kernel, $heap)=@_[KERNEL, HEAP];
     return if $heap->{'is a child'};
@@ -143,7 +153,7 @@ sub server_waste
     
 #------------------------------------------------------------------------------
 # Babysit the child processes
-sub server_babysit
+sub babysit
 {
     my($kernel, $heap)=@_[KERNEL, HEAP];
 
@@ -151,60 +161,153 @@ sub server_babysit
               $heap->{'is a child'};        # or if we are a child
 
     my @children=keys %{$heap->{children}};
-    DEBUG and warn  "$$: Scanning children ", join(", ", sort @children), "\n";
-    my(%missing, $state, $stateT);
-    foreach my $pid (@children) {
-        unless(open STATUS, "/proc/$pid/status") {
-            warn "$$: Unable to open /proc/$pid/status!  Is the child missing: $!\n";
-            $missing{$pid}=1;
-        } else {
-            undef($state);
-            undef($stateT);
-            while(<STATUS>) {
-                next unless /^State:\s*(\w)\s*\((\w+)\)/;
-                ($state, $stateT)=($1,$2);
-                last;
-            }
-            close STATUS or warn $!;
+    $heap->{verbose} and warn  "$$: Babysiting ", scalar(@children), 
+                            " children ", join(", ", sort @children), "\n";
+    my %table;
 
-            if(defined $state) {
-                if($state eq 'Z') {
-                    my $t=waitpid($pid, POSIX::WNOHANG());
-                    if($t==$pid) {
-                        # process was reaped, now fake a SIGCHLD
-                        $kernel->yield('signal', 'CHLD', $pid, $?);
-                        DEBUG and warn "$$: Faking a CHLD for $pid\n";            
-                    } else {
-                        $heap->{verbose} and warn "$$: $pid is a $state ($stateT) and couldn't be reaped.\n";
-                        $missing{$pid}=1;
-                    }
-                } elsif($state eq 'R' or $state eq 'S') {
-                    # do nothing
+    if($heap->{proctable}) {
+        my $table=$heap->{proctable}->table;
+        %table=map {($_->pid, $_)} @$table
+    }
+
+    my(%missing, $state, $time, %rogues, %ok);
+    foreach my $pid (@children) {
+        if($table{$pid}) {
+            $state=$table{$pid}->state;
+
+            if($state eq 'zombie') {
+                my $t=waitpid($pid, POSIX::WNOHANG());
+                if($t==$pid) {
+                    # process was reaped, now fake a SIGCHLD
+                    DEBUG and warn "$$: Faking a CHLD for $pid\n";            
+                    $kernel->yield('sig_CHLD', 'CHLD', $pid, $?, 1);
+                    $ok{$pid}=1;
                 } else {
-                    $heap->{verbose} and warn "$$: $pid has unknown state $state ($stateT)\n";
+                    $heap->{verbose} and warn "$$: $pid is a $state and couldn't be reaped.\n";
+                    $missing{$pid}=1;
                 }
+            } 
+            elsif($state eq 'run') {
+                $time=eval{$table{$pid}->utime + $table{$pid}->stime};
+                warn $@ if $@;
+                # utime and stime are Linux-only :(
+
+                if($time and $time > 600000) { # arbitrary limit of 10 minutes
+                    $rogues{$pid}=$table{$pid};
+                    # DEBUG and 
+                    warn "$$: $pid hass gone rogue, time=$time ms\n";
+                } else {
+                    warn "$$: $pid time=$time ms\n";
+                    $ok{$pid}=1;
+                }
+
+            } elsif($state eq 'sleep') {
+                $ok{$pid}=1;
+                # do nothing
             } else {
-                $heap->{verbose} and warn "$$: Can't find status for $pid!\n";
+                $heap->{verbose} and warn "$$: $pid has unknown state '$state'\n";
+                $ok{$pid}=1;
+            }
+        } elsif($heap->{proctable}) {
+            $heap->{verbose} and warn "$$: $pid isn't in proctable!\n";
+            $missing{$pid}=1;
+        } else {                        # try another means.... :/
+            if(-d "/proc" and not -d "/proc/$pid") {
+                DEBUG and warn "$$: Unable to stat /proc/$pid!  Is the child missing\n";
                 $missing{$pid}=1;
+            } elsif(not $missing{$pid}) {
+                $ok{$pid}=1;
             }
         }
     }
 
+    # if a process is MIA, we fake a death, and spawn a new child
     foreach my $pid (keys %missing) {
-        $kernel->yield('signal', 'CHLD', $pid, 0);
-        DEBUG and warn "$$: Faking a CHLD for $pid MIA\n";            
+        $kernel->yield('sig_CHLD', 'CHLD', $pid, 0, 1);
+        $heap->{verbose} and warn "$$: Faking a CHLD for $pid MIA\n";            
     }
 
-    $kernel->delay('babysit', 60);
+    # we could do the same thing for rogue processes, but instead we
+    # give them time to calm down
+
+    if($heap->{rogues}) {           # processes that are %ok are now removed
+                                    # from the list of rogues
+        delete @{$heap->{rogues}}{keys %ok} if %ok;
+    }
+
+    if(%rogues) {
+        $kernel->yield('rogues') if not $heap->{rogues};
+
+        $heap->{rogues}||={};
+        foreach my $pid (keys %rogues) {
+            if($heap->{rogues}{$pid}) {
+                $heap->{rogues}{$pid}{proc}=$rogues{$pid};
+            } else {
+                $heap->{rogues}{$pid}={proc=>$rogues{$pid}, tries=>0};
+            }
+        }
+    }
+
+    $kernel->delay('babysit', $heap->{babysit});
     return;
 }
 
+#------------------------------------------------------------------------------
+# Deal with rogue child processes
+sub rogues
+{
+    my($kernel, $heap)=@_[KERNEL, HEAP];
+
+    return if $heap->{'die'} or             # don't scan if we are dieing
+              $heap->{'is a child'};        # or if we are a child
+
+                                            # make sure we have some real work
+    return unless $heap->{rogues};
+eval {
+    if(ref($heap->{rogues}) ne 'HASH' or not keys %{$heap->{rogues}}) {
+        delete $heap->{rogues};
+        return;
+    }
+
+    my $signal;
+    while(my($pid, $rogue)=each %{$heap->{rogues}}) {
+        $signal=0;
+        if($rogue->{tries} < 1) {
+            $signal=2;
+        } 
+        elsif($rogue->{tries} < 2) {
+            $signal=15;
+        }
+        elsif($rogue->{tries} < 3) {
+            $signal=9;
+        }
+    
+        if($signal) {
+            DEBUG and warn "$$: Sending signal $signal to rogue $pid\n";
+            unless($rogue->{proc}->kill($signal)) {
+                warn "$$: Error sending signal $signal to $pid: $!\n";
+                delete $heap->{rogues}{$pid};
+            }
+        } else {
+            # if SIGKILL didn't work, it's beyond hope!
+            $kernel->yield('sig_CHLD', 'CHLD', $pid, 0, 1);
+            delete $heap->{rogues}{$pid};
+            $heap->{verbose} and warn "$$: Faking a CHLD for rogue $pid\n";            
+        }
+
+        $rogue->{tries}++;
+    }
+    $kernel->delay('rogues', 2*$heap->{babysit});
+};
+    warn "$$: $@" if $@;
+}
+
 #------------------------------------------------------------------------------
 # Accept POE's standard _stop event, and stop all the children, too.
-# The 'children' hash is maintained in the 'fork' and 'signal'
+# The 'children' hash is maintained in the 'fork' and 'sig_CHLD'
 # handlers.  It's empty for children.
 
-sub server_stop 
+sub _stop 
 {
     my $heap = $_[HEAP];
                                         # kill the child servers
@@ -222,11 +325,11 @@ sub server_stop
 # error occurs while initializing the factory's listening socket, it
 # will exit anyway.
 
-sub server_error 
+sub error 
 {
     my ($heap, $operation, $errnum, $errstr) = @_[HEAP, ARG0, ARG1, ARG2];
     warn __PACKAGE__, " encountered $operation error $errnum: $errstr\n";
-    if($errnum==98) {
+    if($errnum==98) {       # EADDRINUSE
         $heap->{'die'}=1;
     }
 }
@@ -235,19 +338,16 @@ sub server_error
 # The socket factory invokes this state to take care of accepted
 # connections.
 
-sub server_accept 
+sub accept 
 {
     my ($heap, $kernel, $handle, $peer_host, $peer_port) = 
             @_[HEAP, KERNEL, ARG0, ARG1, ARG2];
 
     if(DEBUG) {
-        if ($heap->{'is a child'}) {
             print "$$: Server connection from ", inet_ntoa($peer_host), 
-                            ":$peer_port (Connection $heap->{connections})\n";
-        } else {
-            print "$$: Server connection from ", inet_ntoa($peer_host), 
-                            ":$peer_port\n";
-        }
+                            ":$peer_port", 
+                            ($heap->{'is a child'}  ? 
+                            " (Connection $heap->{connections})\n" : "\n");
     }
     if($heap->{children} and not $heap->{'is a child'}) {
         warn "Parent process received a connection: THIS SUCKS\n";
@@ -275,7 +375,7 @@ sub server_accept
 
 #------------------------------------------------------------------------------
 # The server has been requested to fork, so fork already.
-sub server_fork 
+sub fork 
 {
     my ($kernel, $heap) = @_[KERNEL, HEAP];
     # children should not honor this event
@@ -314,7 +414,8 @@ sub server_fork
     else {
         $heap->{verbose} and warn "$$: Created ", scalar localtime, "\n";
         $heap->{'is a child'}   = 1;        # don't allow fork
-        $heap->{children}       = { };      # don't kill child processes
+        $heap->{'fail forks'}   = 0;
+        $heap->{children}={};               # don't kill child processes
                                             # limit sessions, then die off
         $heap->{connections}    = $heap->{"max connections"};   
 
@@ -334,7 +435,7 @@ sub server_fork
 # Retry failed forks.  This is invoked (after a brief delay) if the
 # 'fork' state encountered a temporary error.
 
-sub server_retry 
+sub retry 
 {
     my ($kernel, $heap) = @_[KERNEL, HEAP];
     if($heap->{'is a child'} or not $heap->{children}) {
@@ -354,51 +455,61 @@ sub server_retry
 }
 
 #------------------------------------------------------------------------------
-# Process signals.  SIGCHLD causes this session to fork off a
-# replacement for the lost child.  Terminal signals aren't handled, so
-# the session will stop on SIGINT.  The _stop event handler takes care
-# of cleanup.
+# SIGCHLD causes this session to fork off a replacement for the lost child.
 
-sub server_signal 
+sub sig_CHLD
 {
-    my ($kernel, $heap, $signal, $pid, $status) =
-                @_[KERNEL, HEAP, ARG0, ARG1, ARG2];
+    my ($kernel, $heap, $signal, $pid, $status, $fake) =
+                @_[KERNEL, HEAP, ARG0, ARG1, ARG2, ARG3];
 
+    return 0 if $heap->{"is a child"};
 
-    return if $heap->{"is a child"};
-
-      # Some operating systems call this SIGCLD.  POE's kernel translates
-      # CLD to CHLD, so developers only need to check for the one version.
-    if($heap->{children} and $signal eq 'CHLD') {
-                                        # if it was one of ours; fork another
+    if($heap->{children}) {
+                                # if it was one of ours; fork another
         if (delete $heap->{children}->{$pid}) {
             DEBUG &&
                     print( "$$: master caught SIGCHLD for $pid.  children: (",
                                 join(' ', sort keys %{$heap->{children}}), ")\n"
                         );
-            $heap->{verbose} and warn "$$: Child $pid exited.\n";
+            $heap->{verbose} and warn "$$: Child $pid ", 
+                        ($fake?'is gone':'exited normaly'), ".\n";
             $kernel->yield('fork') unless $heap->{'die'};
+        } elsif($fake) {
+            warn "$$: Needless fake CHLD for $pid\n";
         } else {
-            warn "$$: CHLD for a child of someone else.\n";
+            warn "$$: CHLD for $pid child of someone else.\n";
         }
     }
+                                        # don't handle terminal signals
+    return 0;
+}
 
-    if($signal eq 'INT') {
-        if($heap->{children}) {
-            warn "$$ SIGINT\n";
-            $heap->{'die'}=1;
-            return 1;
-        } else {
-            delete $heap->{wheel};
-            return 1;
-        }
+#------------------------------------------------------------------------------
+# Terminal signals aren't handled, so the session will stop on SIGINT.  
+# The _stop event handler takes care of cleanup.
+
+sub sig_INT
+{
+    my ($kernel, $heap, $signal, $pid, $status) =
+                @_[KERNEL, HEAP, ARG0, ARG1, ARG2];
+
+    return 0 if $heap->{"is a child"};
+
+    if($heap->{children}) {
+        warn "$$ SIGINT\n";
+        $heap->{'die'}=1;
+        $kernel->delay('waste_time');   # kill this event
+        return 1;
+    } else {
+        delete $heap->{wheel};
+        return 1;
     }    
                                         # don't handle terminal signals
     return 0;
 }
 
 
-sub other_signal
+sub sig_other
 {
 #    return unless DEBUG;
     my ($signal, $pid) = @_[ARG0, ARG1];
@@ -457,12 +568,21 @@ Local kernel name.  This is how we shall "advertise" ourself to foreign
 kernels. It acts as a "kernel alias".  This parameter is temporary, pending
 the addition of true kernel names in the POE core.
 
+=item C<verbose>
+
+Print extra information to STDERR if true.  This allows you to see what
+is going on and potentially trace down problems and stuff.
+
 =item C<processes>
 
 Activates the pre-forking server code.  If set to a positive value, IKC will
 fork processes-1 children.  IKC requests are only serviced by the children. 
 Default is 1 (ie, no forking).
 
+=item C<babysit>
+
+Time, in seconds, between invocations of the babysitter event.
+
 =item C<connections>
 
 Number of connections a child will accept before exiting.  Currently,
